//  

#include<stdio.h>
#include<memory>

#define NODE_SIZE 5	//!<    
#define FREE_SIZE 3	//!<    

inline int min(int a,int b) {if(a>b)a=b;return a;}
inline int max(int a,int b) {if(a<b)a=b;return a;}


class ArrayTree{
protected:
	struct node_st{int array[NODE_SIZE]; node_st *p1; node_st *p2; int count;};
	node_st *root;						//!<  
	void Dump(FILE*,node_st*,int);		//!<    
	void Clear(node_st*);				//!<  
	void FixNode(node_st**);			//!<    
	void GetMin(node_st*,node_st**);	//!<  
	void GetMax(node_st*,node_st**);	//!<       
	node_st *NewNode(int);				//!<   
	bool InsertToNode(node_st*,int&);	//!<  ( )   
	bool Insert(node_st**,int);			//!<    
	bool Remove(node_st**,int);			//!<     
public:
	ArrayTree();
	~ArrayTree();
	bool Insert(int value);		//!<  
	bool Remove(int value);		//!< 
	void Dump(FILE * =stdout);	//!<  
};

ArrayTree::ArrayTree()
{
	root=0;
}

ArrayTree::~ArrayTree()
{
	Clear(root);
}

//!     
void ArrayTree::Clear(node_st *node)
{
	if(!node) return;
	Clear(node->p1);
	Clear(node->p2);
	delete node;
}

//!        
//! \param value    
ArrayTree::node_st *ArrayTree::NewNode(int value)
{
	node_st *node=new node_st;
	node->count=1;
	node->p1=node->p2=0;
	node->array[0]=value;	
	return node;
}

//! \retval true -    
bool ArrayTree::Insert(int value)
{
	return Insert(&root,value);
}

//! \retval true -    
bool ArrayTree::Remove(int value)
{
	return Remove(&root,value);
}

//! \param value 
//! \param    
//! \retval    ,       
//! \retval true -    
bool ArrayTree::Insert(node_st **p,int value)
{
	node_st *node=*p;
	if(!node) {*p=NewNode(value);return false;}
	int n=node->count;
	if(InsertToNode(node,value)) return true;
	if(n<NODE_SIZE) return false;
	if(value<node->array[0]) return Insert(&node->p1,value); 
	else return Insert(&node->p2,value);
}

//! \param value 
//! \param    
//! \retval true      
bool ArrayTree::InsertToNode(node_st *node,int &value)
{
	int n,m,res;
	if(node->count==NODE_SIZE) {
		if(value<node->array[0]) return false;
		if(value>node->array[NODE_SIZE-1]) return false;
	}
	//   
	for(n=0; n<node->count; n++) {
		if(value==node->array[n]) return true;
		if(value<node->array[n]) break;
	}
	if(node->count<NODE_SIZE || rand()&1) {	
		res=node->array[NODE_SIZE-1];
		for(m=NODE_SIZE-1; m>n; m--) node->array[m]=node->array[m-1];
		if(n<NODE_SIZE) node->array[n]=value,value=res; 
	} else {
		res=node->array[0]; n--;
		for(m=0; m<n; m++) node->array[m]=node->array[m+1];
		if(n<0) value=res; else node->array[n]=value,value=res;
	}
	node->count=min(node->count+1,NODE_SIZE);
	return false;
}

//! \param value 
//! \param    
//! \retval    ,      
//! \retval true -    
bool ArrayTree::Remove(node_st **p,int value)
{
	int n;
	node_st *node=*p;
	if(!node) return true;
	if(node->count>=NODE_SIZE) {
		if(value<node->array[0]) return Remove(&node->p1,value);
		if(value>node->array[node->count-1]) return Remove(&node->p2,value);
	}
	for(n=0; n<node->count; n++) if(node->array[n]==value) break;
	if(n==node->count) return true;
	for(;n<node->count-1; n++) node->array[n]=node->array[n+1];
	node->count--;
	FixNode(p);
	return false;
}

//! \param node   
//! \result    ,      
void ArrayTree::FixNode(node_st **p)
{
	node_st *node=*p;
	if(!node->count) {delete node;*p=0;return;}
	if(node->count>=NODE_SIZE-FREE_SIZE) return;
		//     
	if(node->p1 || (node->p1 && node->p2 && rand()&1)) GetMax(node,&node->p1);
	else if(node->p2) GetMin(node,&node->p2);
}

//! \param node   
//! \retval  
//! \result    ,      
void ArrayTree::GetMin(node_st *root,node_st **p)
{
	node_st *node;
	for(;;) {
		node=*p; if(!node->p1) break;
		p=&node->p1;
	}
	int n,offset;
	offset=min(node->count,FREE_SIZE+1);
	node->count-=offset;
	for(n=0; n<offset; n++) root->array[n+root->count]=node->array[n];	// 
	for(n=0; n<node->count; n++) node->array[n]=node->array[n+offset];	// 
	root->count+=offset;
	FixNode(p);
}

//! \param node   
//! \retval  
//! \result    ,      
void ArrayTree::GetMax(node_st *root,node_st **p) 
{
	node_st *node;
	for(;;) {
		node=*p; if(!node->p2) break;
		p=&node->p2;
	}
	int n,count;
	count=min(node->count,FREE_SIZE+1);
	node->count-=count;
	for(n=root->count-1; n>=0; n--) root->array[n+count]=root->array[n];	// 
	for(n=0; n<count; n++) root->array[n]=node->array[node->count+n];		// 
	root->count+=count;
	FixNode(p);
}

void ArrayTree::Dump(FILE *file)
{
	fprintf(file,"[dump]\n");
	Dump(file,root,0);
}

void ArrayTree::Dump(FILE *file,node_st *node,int depth)
{
	if(!node) return;
	Dump(file,node->p1,depth+1);
	printf("%*s[",depth*2,"");
	for(int n=0; n<node->count; n++) {
//		if(node->array[n]==UNDEF_VALUE) break;
		if(n) fputc(' ',file);
		fprintf(file,"%d",node->array[n]);
	}
	fprintf(file,"]\n");
	Dump(file,node->p2,depth+1);
}


void main()
{
	int n,v;
	ArrayTree tree;
	for(n=1; n<=700; n++) {
		v=rand()%100; printf("+[%d] ",v); tree.Insert(v);
		v=rand()%100; printf("-[%d] ",v); tree.Remove(v);
	}
	printf("\n\n\n");
	tree.Dump();
	printf("\n");
	printf("[done]\n");
	getchar();

}
